# ABSTRACT: just main_module for scanrdeps.pl

package App::scanrdeps;
use 5.008001;
use strict;
use warnings;

our $VERSION = "0.01";

=encoding utf-8

=head1 NAME

App::scanrdeps

=head1 INSTALLATION

    cpm install -gv App::scanrdeps

=head1 AUTHOR

Pavel Serikov E<lt>pavelsr@cpan.orgE<gt>

=head1 LICENSE

MIT

=cut

1;

__END__


